from operator import getitem
from Aspidites.tests._vendor.RestrictedPython.helper import restricted_eval

def test_slice():
    rglb = {'_getitem_': getitem}  # restricted globals

    assert restricted_eval('[1, 2, 3, 4, 5]', rglb) == [1, 2, 3, 4, 5]
    assert restricted_eval('[1, 2, 3, 4, 5][:]', rglb) == [1, 2, 3, 4, 5]
    assert restricted_eval('[1, 2, 3, 4, 5][1:]', rglb) == [2, 3, 4, 5]
    assert restricted_eval('[1, 2, 3, 4, 5][:4]', rglb) == [1, 2, 3, 4]
    assert restricted_eval('[1, 2, 3, 4, 5][1:4]', rglb) == [2, 3, 4]
    assert restricted_eval('[1, 2, 3, 4, 5][::3]', rglb) == [1, 4]
    assert restricted_eval('[1, 2, 3, 4, 5][1::3]', rglb) == [2, 5]
    assert restricted_eval('[1, 2, 3, 4, 5][:4:3]', rglb) == [1, 4]
    assert restricted_eval('[1, 2, 3, 4, 5][1:4:3]', rglb) == [2]



from ..interface import Contract, ContractNotRespected, describe_type
from ..syntax import (Keyword, O, S, W, add_contract, add_keyword,
    contract_expression)
import collections


class ASet(Contract):

    def __init__(self, length_contract=None,
                 elements_contract=None, where=None):
        Contract.__init__(self, where)
        self.length_contract = length_contract
        self.elements_contract = elements_contract

    def check_contract(self, context, value, silent):
        if not isinstance(value, collections.Set):
            error = 'Expected a set, got %r.' % describe_type(value)
            raise ContractNotRespected(self, error, value, context)

        if self.length_contract is not None:
            self.length_contract._check_contract(context, len(value), silent)

        if self.elements_contract is not None:
            for element in value:
                self.elements_contract._check_contract(context, element, silent)

    def __str__(self):
        s = 'set'
        if self.length_contract is not None:
            s += '[%s]' % self.length_contract
        if self.elements_contract is not None:
            s += '(%s)' % self.elements_contract
        return s

    def __repr__(self):
        s = 'ASet(%r,%r)' % (self.length_contract, self.elements_contract)
        return s

    @staticmethod
    def parse_action(s, loc, tokens):
        where = W(s, loc)
        length_contract = tokens.get('length_contract', None)
        elements_contract = tokens.get('elements_contract', None)
        return ASet(length_contract, elements_contract, where=where)


list_contract = (Keyword('set') -
                 O(S('[') - contract_expression('length_contract') - S(']')) +
                 O(S('(') - contract_expression('elements_contract') - S(')')))
list_contract.setParseAction(ASet.parse_action)

list_contract.setName('Set contract')
add_keyword('set')
add_contract(list_contract)




# Aspidites
# Copyright (C) 2021 Ross J. Duff

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import os
import py_compile
import sys
from glob import glob
import typing as t
from .templates import lib, makefile, pyproject, setup, default_file, inject_stack
from pyrsistent import pmap, v
from hashlib import sha256
from pathlib import Path
from .final import final
from ._vendor.semantic_version import Version
from pyparsing import ParseResults


@final()
class CheckedFileStack:

    """A convenience class for reading file data streams to stdout or to checksum"""

    __slots__ = v('all_files', 'pre_size')

    def __init__(self, initial=None, pre_size=128):
        if initial is None:
            initial = {}
        self.all_files = pmap(initial, pre_size)
        self.all_files = self.all_files.evolver()
        self.pre_size = pre_size

    def __repr__(self):
        return self.__class__.__name__

    def _read(self, data, print__=False, hash_func=None):
        chunk = data.read(self.pre_size)
        if hash_func is None:
            curr_hash = hash_func
        else:
            curr_hash = hash_func()
        while chunk:
            curr_hash and curr_hash.update(chunk)  # Short-circuits if called without hash_func
            chunk = data.read(self.pre_size)
            if print__:
                print(chunk, '')
        return curr_hash

    def _write_checksum(self, fname) -> tuple:
        fname = Path(fname)
        base, name = fname.parent, fname.name
        base = Path(base)
        fname_sha256 = base / ("." + name + ".sha256")
        with open(fname, "rb") as data:
            curr_hash = self._read(data, hash_func=sha256)
            with open(fname_sha256, "wb") as digest:
                digest.write(curr_hash.digest())
            return pmap({curr_hash.digest(): fname}).items()[0]  # immutable

    def _read_checksum(self, fname):
        fname = Path(fname)
        base, name = fname.parent, fname.name
        base = Path(base)
        fname_sha256 = base / ("." + name + ".sha256")
        old = open(fname_sha256, "rb").read()
        with open(fname, "rb") as data:
            curr_hash = self._read(data, hash_func=sha256)
            new = curr_hash.digest()
            if new == old:
                print(
                    "sha256 digest check successful: %s %s"
                    % (name, new.hex())
                )
                return new
            else:
                print(
                    "sha256 digest failure: %s %s"
                    % (name, new.hex())
                )
                return ""

    def _commit(self):
        """Commits all registered files making the all_files attribute immutable.
        Short-circuits if no files are in the stack."""
        return not not len(self.all_files) and self.all_files.persistent()

    def register(self, fname):
        """Registers a filename to a checksum of its contents."""
        self.all_files.set(*self._write_checksum(fname))

    def create_file(self, fname, mode, root='', text=default_file) -> None:
        """API for creating and registering checked files"""
        if len(str(root)) > 0:
            root = Path(root)
            file = root/fname
        else:
            file = fname
        try:
            open(file, mode).write(text)
        except FileExistsError:
            self.register(file)
        except FileNotFoundError:
            open(file, 'x').write(text)

    def finalize(self) -> None:
        """Read and check all files against their stored digests."""
        all_file_checksums = self._commit()
        if not all_file_checksums:
            self.all_files.persistent()
            return None
        print("running checksums")
        for k, v in all_file_checksums.items():
            digest = self._read_checksum(v)
            try:
                all_file_checksums.get(digest)
            except AttributeError:
                raise RuntimeError("\nfor file %s\n%s\n  did not match cached digest\n%s")


@final()
class CompilerArgs:
    def __init__(self, **kwargs):
        self.code: ParseResults = kwargs['code']
        self.fname: Path = kwargs['fname']
        self.force: bool = kwargs['force']
        self.bytecode: bool = kwargs['bytecode']
        self.c: bool = kwargs['c']
        self.build_requires: t.Union[t.List, str] = kwargs['build_requires']
        self.verbose: int = kwargs['verbose']
        self.embed: t.Union[str, None] = kwargs['embed']
        self.__setattr__ = lambda x, y: None

    def __repr__(self):
        return self.__class__.__name__


class Compiler:
    def __init__(self, **kwargs):
        self.args = CompilerArgs(**kwargs)
        self.file_stack = CheckedFileStack()
        self.fname = Path(self.args.fname)
        self.app_name = self.fname.parent / self.fname.stem
        self.project = self.app_name.stem
        self.module_name = str(self.app_name).replace("/", ".")
        self.file_c = str(self.app_name) + ".c"
        self.root = self.fname.parent
        self.mode = "x" if self.args.force else "w"
        files = {
            self.fname: (
                self.mode,
                dict(root='', text=lib.substitute(code="\n".join(self.args.code)))
            ),
            '__init__.py': (
                self.mode,
                dict(root=self.root, text=f'__metadata__ = "{self.__dict__}"')
            ),
            'py.typed': (
                self.mode,
                dict(root=self.root)
            ),
            'pyproject.toml': (
                self.mode,
                dict(root=self.root, text=pyproject.substitute(build_requires=self.args.build_requires))
            ),
            'Makefile': (
                self.mode,
                dict(root=self.root, text=makefile.substitute(project=self.project))
            ),
        }

        for k, v in files.items():
            args_, kwargs_ = v
            self.file_stack.create_file(k, args_, **kwargs_)

        if self.args.bytecode:
            self.bytecode_compile()

        if self.args.c:
            # self.compile_c()
            if self.args.embed and 'main' in self.args.embed:
                pass  # maybe write a wrapper or something idk?
            self.setup(**kwargs)

        self.file_stack.finalize()

    def bytecode_compile(self):
        fname_pyc = str(self.app_name) + ".pyc"
        quiet = tuple(reversed(range(3))).index(self.args.verbose if self.args.verbose < 2 else 2)
        major, minor, patch, *_ = sys.version_info
        if Version(major=major, minor=minor, patch=patch) < Version('3.8.0'):
            py_compile.compile(str(self.fname), fname_pyc)
        else:
            py_compile.compile(str(self.fname), fname_pyc, quiet=quiet)
        self.file_stack.register(fname_pyc)

    def setup(self, **kwargs) -> None:
        module_name = str(self.app_name).replace("/", ".")
        text = setup.substitute(
           app_name=module_name,
           src_file=kwargs['fname'],
           inc_dirs=[],
           libs=[],
           exe_name=self.app_name,
           lib_dirs=[],
           **kwargs
        )
        self.file_stack.create_file('setup.py', self.mode, root=str(self.root), text=text)
        self.compile_object()

    def compile_object(self) -> None:
        glob_so = str(self.app_name) + ".*.so"
        # TODO: get this working for docker builds
        #  (maybe executable param with os.path.relpath?)
        setup_runner = "%s %s build_ext -b ." % (sys.executable, str(Path(self.root) / 'setup.py'))
        print("running", setup_runner)
        with os.popen(setup_runner) as p:
            self.file_stack._read(p, print__=True)
        self.file_stack.register(self.file_c)
        for i in glob(glob_so):
            self.file_stack.register(i)



import sys
import inspect
from .backported import getcallargs, getfullargspec
from ..fn.underscore import ArityError

inPy2 = sys.version_info[0] == 2
if inPy2:  # pragma: no cover
    from types import ClassType


def can_be_used_as_a_type(x):
    """ Checks that x can be used as a type; specifically,
        we can write isintance(y,x).
        Here we support old-style classes.
    """
    if isinstance(x, type):
        return True

    if inPy2:  # pragma: no cover
        if isinstance(x, ClassType):
            return True

    return False


def can_accept_exactly_one_argument(callable_thing):
    """ Checks that a callable can accept exactly one argument
        using introspection.
    """
    if inspect.ismethod(callable_thing):  # bound method
        f = callable_thing.__func__
        args = (callable_thing.__self__, 'test',)
    else:
        if not inspect.isfunction(callable_thing):
            f = callable_thing.__call__
        else:
            f = callable_thing
        args = ('test',)

    try:
        getcallargs(f, *args)
    except (TypeError, ValueError, ArityError) as e:  # @UnusedVariable
        # print 'Get call args exception (f=%r,args=%r): %s ' % (f, args, e)
        return False, str(e)
    else:
        return True, None


def get_f_from_callable(callable_thing):
    if inspect.ismethod(callable_thing):  # bound method
        f = callable_thing.__func__
        # args = (callable_thing.__self__, 'test',)
    else:
        if not inspect.isfunction(callable_thing):
            f = callable_thing.__call__
        else:
            f = callable_thing
            # args = ('test',)
    return f


def get_callable_fullargspec(callable_thing):
    f = get_f_from_callable(callable_thing)
    spec = getfullargspec(f)
    return spec


def can_accept_at_least_one_argument(callable_thing):
    """ 
        Checks that a callable can accept at least one argument
        using introspection.
    """
    spec = get_callable_fullargspec(callable_thing)
    return len(spec.args) > 0 or spec.varargs


def can_accept_self(callable_thing):
    """ 
        Checks that a callable's first argument is self
    """
    spec = get_callable_fullargspec(callable_thing)

    if len(spec.args) == 0 or spec.args[0] != 'self':
        return False

    return True


def can_accept_self_plus_one_argument(callable_thing):
    """ 
        Checks that a callable can accept exactly self plus one argument
        using introspection.
    """

    spec = get_callable_fullargspec(callable_thing)
    if len(spec.args) == 0 or spec.args[0] != 'self':
        return False

    # TODO: redo better
    f = get_f_from_callable(callable_thing)
    try:
        getcallargs(f, 'self', 'value')
    except (TypeError, ValueError, ArityError) as e:  # @UnusedVariable
        return False
    else:
        return True


class InvalidArgs(Exception):
    pass

def check_callable_accepts_these_arguments(callable_thing, args, kwargs):
    """ Checks that a callable can accept the args and kwargs. 
    
        Returns either None or raises InvalidArgs. 
    """
    f = get_f_from_callable(callable_thing)

    # TODO: more cleanly
    try:
        bound = getcallargs(f, *args, **kwargs)
        # print('bound: %r ' % bound)
    except (TypeError, ValueError) as e:  # @UnusedVariable
        # print('no!: %s' % e)
        raise InvalidArgs('%s does not accept %s, %s: %s' % (f, args, kwargs, e))
        # return False
    else:
        return True

    return False



from typing import Any

safe_builtins: Any
full_write_guard: Any

def guarded_setattr(object, name, value) -> None: ...
def guarded_delattr(object, name) -> None: ...
def safer_getattr(object, name, default: Any | None = ..., getattr=...): ...
def guarded_iter_unpack_sequence(it, spec, _getiter_) -> None: ...
def guarded_unpack_sequence(it, spec, _getiter_): ...

safe_globals: Any



import sys
from warnings import warn
from inspect import isfunction, signature, getouterframes
from cmath import inf, isinf, nan, isnan
import numbers
from pyrsistent import v, pvector
from .templates import _warning
from .api import bordered


# noinspection PyPep8Naming,PyProtectedMember,PyUnresolvedReferences
def SafeUnaryAdd(a):
    if isnan(a) or not isinstance(a, numbers.Number):
        stack = pvector(getouterframes(sys._getframe(0), 1))
        exc = ZeroDivisionError("Unary Add is Undefined for %s" % type(a))
        w = Warn(stack, stack[0][3], [a], {}).create(exc)
        warn(w, category=RuntimeWarning, stacklevel=0)
        return Undefined()
    return +a


# noinspection PyPep8Naming,PyProtectedMember,PyUnresolvedReferences
def SafeUnarySub(a):
    if isnan(a) or not isinstance(a, numbers.Number):
        stack = pvector(getouterframes(sys._getframe(0), 1))
        exc = ZeroDivisionError("Unary Sub is Undefined for %s" % type(a))
        w = Warn(stack, stack[0][3], [a], {}).create(exc)
        warn(w, category=RuntimeWarning, stacklevel=0)
        return Undefined()
    return -a


# noinspection PyPep8Naming,PyProtectedMember,PyUnresolvedReferences
def SafeFloorDiv(a, b):
    """IEEE 754-1985 evaluates an expression and replaces indeterminate forms with Undefined instances"""
    if isinf(a) or b == 0 or (isinf(a) and isinf(b)):
        stack = pvector(getouterframes(sys._getframe(0), 1))
        exc = ZeroDivisionError("Division by zero is Undefined; this behavior diverges from IEEE 754-1985.")
        w = Warn(stack, stack[0][3], [a, b], {}).create(exc)
        warn(w, category=RuntimeWarning, stacklevel=0)
        return Undefined()
    return a // b


# noinspection PyPep8Naming,PyProtectedMember,PyUnresolvedReferences
def SafeDiv(a, b):
    """IEEE 754-1985 evaluates an expression and replaces indeterminate forms with Undefined instances"""
    if b == 0 or (isinf(a) and isinf(b)):
        stack = pvector(getouterframes(sys._getframe(0), 1))
        exc = ZeroDivisionError("Division by zero is Undefined; this behavior diverges from IEEE 754-1985.")
        w = Warn(stack, stack[0][3], [a, b], {}).create(exc)
        warn(w, category=RuntimeWarning, stacklevel=0)
        return Undefined()
    return a / b


# noinspection PyPep8Naming, PyProtectedMember,PyUnresolvedReferences
def SafeMod(a, b):
    """IEEE 754-1985 evaluates an expression and replaces indeterminate forms with Undefined instances"""
    if isinf(a) or b == 0:
        stack = pvector(getouterframes(sys._getframe(0), 1))
        exc = ZeroDivisionError("Modulus by zero is Undefined; this behavior diverges from IEEE 754-1985.")
        w = Warn(stack, stack[0][3], [a, b], {}).create(exc)
        warn(w, category=RuntimeWarning, stacklevel=0)
        return Undefined()
    return a % b


# noinspection PyPep8Naming, PyProtectedMember,PyUnresolvedReferences
def SafeExp(a, b):
    if (a == 0 and b == 0) or (isinf(a) and b == 0) or (isinf(b) and a == 0):  # 0**0, inf**0, 0**inf
        stack = pvector(getouterframes(sys._getframe(0), 1))
        exc = ArithmeticError(
            "%s**%s" % (str(a), str(b),) + " == Undefined; this behavior diverges from IEEE 754-1985."
        )
        w = Warn(stack, stack[0][3], [a, b], {}).create(exc)
        warn(w, category=RuntimeWarning, stacklevel=0)
        return Undefined()
    try:
        return a**b
    except OverflowError:
        return inf  # just a really big number on most systems


class Warn:

    def __init__(self, stack, func, *args, **kwargs):
        self.stack = stack
        self.func = func
        self.args = args
        self.kwargs = kwargs

    def create(self, exc=Exception()):
        _locals = self.stack[1][0].f_locals.items()
        str_locals = self.format_locals(_locals, exc)
        func_name = self.stack[1][0].f_code.co_name
        fname = self.stack[1][0].f_code.co_filename
        lineno = self.stack[1][0].f_code.co_firstlineno
        fkwargs = self.format_kwargs()
        if hasattr(self.func, "__name__"):
            name = self.func.__name__
        else:
            name = str(self.func)
        atfault = (
            name
            if isinstance(exc, TypeError)
            else name + "(" + str(self.args).strip("()") + fkwargs + ")"
        )
        return _warning.safe_substitute(
            file=fname,
            lineno=lineno,
            func=bordered(func_name),
            atfault=bordered(atfault),
            bound=bordered(str_locals),
            tb=bordered(str(exc)),
        )

    def format_kwargs(self, sep: str = ", "):
        return sep + str(self.kwargs).strip("{} ").replace(":", "=") if len(self.kwargs) else ""

    # noinspection PyMethodMayBeStatic
    def format_locals(self, local_vars, exc: Exception):
        locals_ = dict(filter(lambda x: x[1] != str(exc), local_vars))
        str_locals = str()
        for k, v_ in locals_.items():
            if str(k).startswith("@"):  # skip @py_assert
                continue
            if isfunction(v_):
                str_locals += k + ": " + str(signature(v_)).replace("'", "") + "\n"
            else:
                str_locals += k + ": " + str(v_) + "\n"
        return str_locals.rstrip("\n")


class Undefined:
    """A monad for a failed programmatic unit; like NoneType but hashable.
    Falsy singleton"""

    __slots__ = v("__weakref__", "__instance__")
    __instance = None

    def __hash__(self):
        # noinspection PyUnresolvedReferences
        return hash(self.__weakref__)

    def __eq__(self, other):
        return self.__hash__ == other.__hash__

    def __add__(self, other):
        return self

    def __sub__(self, other):
        return self

    def __neg__(self):
        return self

    def __invert__(self):
        return self

    def __mul__(self, other):
        return self

    def __truediv__(self, other):
        return self

    def __floordiv__(self, other):
        return self

    def __str__(self):
        return self.__repr__()

    def __float__(self):
        return nan

    def __complex__(self):
        return complex(nan)

    def __oct__(self):
        return self

    def __index__(self):
        return 0

    def __len__(self):
        # Undefined has 0 elements
        return 0

    def __repr__(self):
        return self.__class__.__name__ + "()"

    # noinspection PyMethodMayBeStatic
    def __nonzero__(self):
        return True

    def __call__(self, *args, **kwargs):
        return self.__new__(self.__class__, *args, **kwargs)

    # noinspection PyMethodParameters
    def __new__(mcs, *args, **kwargs):
        if mcs.__instance is None:
            mcs.__instance = super(Undefined, mcs).__new__(mcs, *args, **kwargs)
            mcs.__instance__ = mcs.__instance
        # noinspection PyUnresolvedReferences
        return mcs.__instance__  # instance descriptor from __slots__ -> actual instance



import operator
import random
import re
import string
from itertools import count, repeat
from sys import version_info

from .func import F
from .op import apply, flip, identity


div = operator.truediv

letters = string.ascii_letters


def _random_name():
    return "".join(random.choice(letters) for _ in range(14))


def fmap(f, format):
    def applyier(self, other):
        fmt = "(%s)" % format.replace("self", self._format)

        if isinstance(other, self.__class__):
            return self.__class__(
                (f, self, other),
                fmt.replace("other", other._format),
                dict(
                    list(self._format_args.items()) +
                    list(other._format_args.items())
                ),
                self._arity + other._arity
            )
        else:
            call = F(flip(f), other) << F(self)
            name = _random_name()
            return self.__class__(
                call,
                fmt.replace("other", "%%(%s)r" % name),
                dict(list(self._format_args.items()) + [(name, other)]),
                self._arity)
    return applyier


class ArityError(TypeError):
    def __str__(self):
        return "{0!r} expected {1} arguments, got {2}".format(*self.args)


def unary_fmap(f, format):
    def applyier(self):
        fmt = "(%s)" % format.replace("self", self._format)
        return self.__class__(
            F(self) << f, fmt, self._format_args, self._arity
        )
    return applyier


class _Callable(object):

    __slots__ = "_callback", "_format", "_format_args", "_arity"
    # Do not use "flipback" approach for underscore callable,
    # see https://github.com/kachayev/fn.py/issues/23
    __flipback__ = None

    def __init__(self, callback=identity, format="_", format_args=None,
                 arity=1):
        self._callback = callback
        self._format = format
        self._format_args = format_args or {}
        self._arity = arity

    def call(self, name, *args, **kwargs):
        """Call method from _ object by given name and arguments"""
        return self.__class__(
            F(lambda f:
              apply(f, args, kwargs)) << operator.attrgetter(name) << F(self)
        )

    def __getattr__(self, name):
        if name == '__wrapped__':  # Guard for recursive call by doctest
            raise AttributeError
        attr_name = _random_name()
        return self.__class__(
            F(operator.attrgetter(name)) << F(self),
            "getattr(%s, %%(%s)r)" % (self._format, attr_name),
            dict(
                list(self._format_args.items()) + [(attr_name, name)]
            ),
            self._arity
        )

    def __getitem__(self, k):
        if isinstance(k, self.__class__):
            return self.__class__(
                (operator.getitem, self, k),
                "%s[%s]" % (self._format, k._format),
                dict(
                    list(self._format_args.items()) +
                    list(k._format_args.items())
                ),
                self._arity + k._arity
            )
        item_name = _random_name()
        return self.__class__(
            F(operator.itemgetter(k)) << F(self),
            "%s[%%(%s)r]" % (self._format, item_name),
            dict(list(self._format_args.items()) + [(item_name, k)]),
            self._arity
        )

    def __str__(self):
        """Build readable representation for function

        (_ < 7): (x1) => (x1 < 7)
        (_ + _*10): (x1, x2) => (x1 + (x2*10))
        """
        # args iterator with produce infinite sequence
        # args -> (x1, x2, x3, ...)
        args = map("".join, zip(repeat("x"), map(str, count(1))))
        l, r = [], self._format
        # replace all "_" signs from left to right side
        while r.count("_"):
            n = next(args)
            r = r.replace("_", n, 1)
            l.append(n)

        r = r % self._format_args
        return "({left}) => {right}".format(left=", ".join(l), right=r)

    def __repr__(self):
        """
        Return original function notation to ensure that eval(repr(f)) == f
        """
        return re.sub(r"x\d+", "_", str(self).split("=>", 1)[1].strip())

    def __call__(self, *args):
        if len(args) != self._arity:
            raise ArityError(self, self._arity, len(args))

        if not isinstance(self._callback, tuple):
            return self._callback(*args)

        f, left, right = self._callback
        return f(left(*args[:left._arity]), right(*args[left._arity:]))

    __add__ = fmap(operator.add, "self + other")
    __mul__ = fmap(operator.mul, "self * other")
    __sub__ = fmap(operator.sub, "self - other")
    __mod__ = fmap(operator.mod, "self %% other")
    __pow__ = fmap(operator.pow, "self ** other")

    __and__ = fmap(operator.and_, "self & other")
    __or__ = fmap(operator.or_, "self | other")
    __xor__ = fmap(operator.xor, "self ^ other")

    __div__ = fmap(div, "self / other")
    __divmod__ = fmap(divmod, "self / other")
    __floordiv__ = fmap(operator.floordiv, "self / other")
    __truediv__ = fmap(operator.truediv, "self / other")

    __lshift__ = fmap(operator.lshift, "self << other")
    __rshift__ = fmap(operator.rshift, "self >> other")

    __lt__ = fmap(operator.lt, "self < other")
    __le__ = fmap(operator.le, "self <= other")
    __gt__ = fmap(operator.gt, "self > other")
    __ge__ = fmap(operator.ge, "self >= other")
    __eq__ = fmap(operator.eq, "self == other")
    __ne__ = fmap(operator.ne, "self != other")

    __neg__ = unary_fmap(operator.neg, "-self")
    __pos__ = unary_fmap(operator.pos, "+self")
    __invert__ = unary_fmap(operator.invert, "~self")

    __radd__ = fmap(flip(operator.add), "other + self")
    __rmul__ = fmap(flip(operator.mul), "other * self")
    __rsub__ = fmap(flip(operator.sub), "other - self")
    __rmod__ = fmap(flip(operator.mod), "other %% self")
    __rpow__ = fmap(flip(operator.pow), "other ** self")
    __rdiv__ = fmap(flip(div), "other / self")
    __rdivmod__ = fmap(flip(divmod), "other / self")
    __rtruediv__ = fmap(flip(operator.truediv), "other / self")
    __rfloordiv__ = fmap(flip(operator.floordiv), "other / self")

    __rlshift__ = fmap(flip(operator.lshift), "other << self")
    __rrshift__ = fmap(flip(operator.rshift), "other >> self")

    __rand__ = fmap(flip(operator.and_), "other & self")
    __ror__ = fmap(flip(operator.or_), "other | self")
    __rxor__ = fmap(flip(operator.xor), "other ^ self")


shortcut = _Callable()



from ..interface import Contract
from ..syntax import add_contract, W, contract_expression, Literal, Group


class SeparateContext(Contract):

    def __init__(self, contract, where=None):
        assert isinstance(contract, Contract)
        Contract.__init__(self, where)
        self.contract = contract

    def check_contract(self, context, value, silent):
        copy = context.copy()
        self.contract._check_contract(copy, value, silent)

    def __str__(self):
        return '$(%s)' % self.contract

    def __repr__(self):
        return 'SeparateContext(%r)' % self.contract

    @staticmethod
    def parse_action(s, loc, tokens):
        where = W(s, loc)
        return SeparateContext(tokens[0]['child'], where=where)


sepcon = (Group(Literal('$') + Literal('(') -
                contract_expression('child') - Literal(')')))
sepcon.setParseAction(SeparateContext.parse_action)
sepcon.setName('Context separation construct')
add_contract(sepcon)



from Aspidites.tests._vendor.RestrictedPython.helper import restricted_eval

def test_UAdd():
    assert restricted_eval('+a', {'a': 42}) == 42


def test_USub():
    assert restricted_eval('-a', {'a': 2411}) == -2411



from Aspidites._vendor.contracts import new_contract

try:
    import numpy
except ImportError:  # pragma: no cover
    pass
else:
    from .numpy_specific import *    

from .numbers import *


new_contract('bytes', bytes)



try:
    import numpy
except ImportError:
    pass
else:

    import unittest

    from Aspidites._vendor.contracts import decorate, new_contract, ContractNotRespected

    new_contract('rgb', 'array[HxWx3],H>0,W>0')
    new_contract('rgba', 'array[HxWx4],H>0,W>0')

    def blend_function(image1, image2, bug=False):
        """
            Blends two RGB or RGBA images together.

             :param image1: The first image to blend.
             :type image1: (rgb|rgba),array[HxWx*]
             :param image2: The second image to blend.
             :type image2: (rgb|rgba),array[HxWx*]
             :param bug: Introduce a bug to check the contracts.
             :type bug: bool

             :return: The blended image.
             :rtype: rgb,array[HxWx3]
        """
        H, W = image1.shape[0], image1.shape[1]

        if bug:
            # if we want to show a bug, return a different shape
            W += 1

        result = numpy.zeros((H, W, 3), 'uint8')

        # put here the actual function
        image2

        return result

    im_float = numpy.zeros((10, 10, 3), dtype='float32')
    rgb_small = numpy.zeros((10, 10, 3), dtype='uint8')
    rgb_large = numpy.zeros((20, 20, 3), dtype='uint8')

    rgba_small = numpy.zeros((10, 10, 3), dtype='uint8')
    rgba_large = numpy.zeros((20, 20, 3), dtype='uint8')

    class ArrayTest(unittest.TestCase):

        def setUp(self):
            self.blend = decorate(blend_function)

        def test_correct_behavior(self):
            self.blend(rgb_small, rgb_small)
            self.blend(rgb_small, rgba_small)
            self.blend(rgba_small, rgb_small)
            self.blend(rgb_large, rgba_large)
            self.blend(rgba_large, rgb_large)

        def test_incorrect1(self):
            self.assertRaises(ContractNotRespected, self.blend, None, None)

        def test_incorrect2(self):
            self.assertRaises(ContractNotRespected, self.blend,
                              None, rgb_small)

        def test_incorrect3(self):
            self.assertRaises(ContractNotRespected, self.blend,
                              rgb_small, None)

        def test_incorrect4(self):
            self.assertRaises(ContractNotRespected, self.blend,
                              rgb_small, rgb_large)

        def test_incorrect5(self):
            self.assertRaises(ContractNotRespected, self.blend,
                              rgb_small, rgb_large)

        def test_incorrect6(self):
            # check that rtype checking works, introduce a bug
            self.assertRaises(ContractNotRespected, self.blend, rgb_small,
                              rgb_small, bug=True)




from Aspidites._vendor.RestrictedPython import compile_restricted
from Aspidites._vendor.RestrictedPython import compile_restricted_eval
from Aspidites._vendor.RestrictedPython import compile_restricted_exec
from Aspidites._vendor.RestrictedPython import compile_restricted_single
from Aspidites._vendor.RestrictedPython import CompileResult
from Aspidites._vendor._compat import IS_PY2
from Aspidites._vendor._compat import IS_PY3
from Aspidites._vendor._compat import IS_PY38_OR_GREATER
from Aspidites.tests._vendor.RestrictedPython.helper import restricted_eval

import platform
import pytest
import types


def test_compile__compile_restricted_invalid_code_input():
    with pytest.raises(TypeError):
        compile_restricted(object(), '<string>', 'exec')
    with pytest.raises(TypeError):
        compile_restricted(object(), '<string>', 'eval')
    with pytest.raises(TypeError):
        compile_restricted(object(), '<string>', 'single')


def test_compile__compile_restricted_invalid_policy_input():
    with pytest.raises(TypeError):
        compile_restricted("pass", '<string>', 'exec', policy=object)


def test_compile__compile_restricted_invalid_mode_input():
    with pytest.raises(TypeError):
        compile_restricted("pass", '<string>', 'invalid')


INVALID_ASSINGMENT = """
1 = 2
"""


def test_compile__invalid_syntax():
    with pytest.raises(SyntaxError) as err:
        compile_restricted(INVALID_ASSINGMENT, '<string>', 'exec')
    if IS_PY38_OR_GREATER:
        assert "cannot assign to literal at statement:" in str(err.value)
    else:
        assert "can't assign to literal at statement:" in str(err.value)


def test_compile__compile_restricted_exec__1():
    """It returns a CompileResult on success."""
    result = compile_restricted_exec('a = 42')
    assert result.__class__ == CompileResult
    assert result.errors == ()
    assert result.warnings == []
    assert result.used_names == {}
    glob = {}
    exec(result.code, glob)
    assert glob['a'] == 42


def test_compile__compile_restricted_exec__2():
    """It compiles without restrictions if there is no policy."""
    result = compile_restricted_exec('_a = 42', policy=None)
    assert result.errors == ()
    assert result.warnings == []
    assert result.used_names == {}
    glob = {}
    exec(result.code, glob)
    assert glob['_a'] == 42


def test_compile__compile_restricted_exec__3():
    """It returns a tuple of errors if the code is not allowed.

    There is no code in this case.
    """
    result = compile_restricted_exec('_a = 42\n_b = 43')
    errors = (
        'Line 1: "_a" is an invalid variable name because it starts with "_"',
        'Line 2: "_b" is an invalid variable name because it starts with "_"')
    assert result.errors == errors
    assert result.warnings == []
    assert result.used_names == {}
    assert result.code is None


def test_compile__compile_restricted_exec__4():
    """It does not return code on a SyntaxError."""
    result = compile_restricted_exec('asdf|')
    assert result.code is None
    assert result.warnings == []
    assert result.used_names == {}
    assert result.errors == (
        "Line 1: SyntaxError: invalid syntax at statement: 'asdf|'",)


def test_compile__compile_restricted_exec__5():
    """It does not return code if the code contains a NULL byte."""
    result = compile_restricted_exec('a = 5\x00')
    assert result.code is None
    assert result.warnings == []
    assert result.used_names == {}
    if IS_PY2:  # pragma: no cover
        assert result.errors == (
            'compile() expected string without null bytes',)
    else:  # pragma: PY3
        assert result.errors == (
            'source code string cannot contain null bytes',)


EXEC_STATEMENT = """\
def no_exec():
    exec 'q = 1'
"""


@pytest.mark.skipif(
    IS_PY2,
    reason="exec statement in Python 2 is handled by Aspidites._vendor.RestrictedPython ")
def test_compile__compile_restricted_exec__10():  # pragma: PY3
    """It is a SyntaxError to use the `exec` statement. (Python 3 only)"""
    result = compile_restricted_exec(EXEC_STATEMENT)
    assert (
        'Line 2: SyntaxError: Missing parentheses in call to \'exec\' at '
        'statement: "exec \'q = 1\'"',) == result.errors


FUNCTION_DEF = """\
def a():
    pass
"""


def test_compile__compile_restricted_eval__1():
    """It compiles code as an Expression.

    Function definitions are not allowed in Expressions.
    """
    result = compile_restricted_eval(FUNCTION_DEF)
    assert result.errors == (
        "Line 1: SyntaxError: invalid syntax at statement: 'def a():'",)


def test_compile__compile_restricted_eval__2():
    """It compiles code as an Expression."""
    assert restricted_eval('4 * 6') == 24


def test_compile__compile_restricted_eval__used_names():
    result = compile_restricted_eval("a + b + func(x)")
    assert result.errors == ()
    assert result.warnings == []
    assert result.used_names == {'a': True, 'b': True, 'x': True, 'func': True}


def test_compile__compile_restricted_csingle():
    """It compiles code as an Interactive."""
    result = compile_restricted_single('4 * 6')
    assert result.code is None
    assert result.errors == (
        'Line None: Interactive statements are not allowed.',
    )


PRINT_EXAMPLE = """
def a():
    print 'Hello World!'
"""


@pytest.mark.skipif(
    IS_PY3,
    reason="Print statement is gone in Python 3."
           "Test Deprecation Warming in Python 2")
def test_compile_restricted():  # pragma: no cover
    """This test checks compile_restricted itself if that emit Python warnings.
    For actual tests for print statement see: test_print_stmt.py
    """
    with pytest.warns(SyntaxWarning) as record:
        result = compile_restricted(PRINT_EXAMPLE, '<string>', 'exec')
        assert isinstance(result, types.CodeType)
        # Non-CPython versions have a RuntimeWarning, too.
        if len(record) > 2:  # pragma: no cover
            record.pop()
        assert len(record) == 1
        assert record[0].message.args[0] == \
            "Line 2: Prints, but never reads 'printed' variable."


EVAL_EXAMPLE = """
def a():
    eval('2 + 2')
"""


def test_compile_restricted_eval():
    """This test checks compile_restricted itself if that raise Python errors.
    """
    with pytest.raises(SyntaxError,
                       match="Line 3: Eval calls are not allowed."):
        compile_restricted(EVAL_EXAMPLE, '<string>', 'exec')


def test_compile___compile_restricted_mode__1(recwarn, mocker):
    """It warns when using another Python implementation than CPython."""
    if platform.python_implementation() == 'CPython':  # pragma: no cover
        # Using CPython we have to fake the check:
        mocker.patch('Aspidites._vendor.RestrictedPython.compile.IS_CPYTHON', new=False)
    with pytest.warns(RuntimeWarning) as record:
        compile_restricted('42')
    assert len(record) == 1
    w = record.pop()
    assert w.category == RuntimeWarning
    assert str(w.message) == str(
        'RestrictedPython is only supported on CPython: use on other Python '
        'implementations may create security issues.'
    )


@pytest.mark.skipif(
    platform.python_implementation() == 'CPython',
    reason='Warning only present if not CPython.')
def test_compile_CPython_warning(recwarn, mocker):  # pragma: no cover
    """It warns when using another Python implementation than CPython."""
    assert platform.python_implementation() != 'CPython'
    with pytest.warns(RuntimeWarning) as record:
        compile_restricted('42')
    assert len(record) == 1
    assert str(record[0].message) == str(
        'RestrictedPython is only supported on CPython: use on other Python '
        'implementations may create security issues.'
    )



from Aspidites._vendor.RestrictedPython import compile_restricted_exec
from Aspidites._vendor._compat import IS_PY2
from Aspidites.tests._vendor.RestrictedPython.helper import restricted_exec

import pytest


BAD_NAME_STARTING_WITH_UNDERSCORE = """\
def bad_name():
    __ = 12
"""


def test_RestrictingNodeTransformer__visit_Name__1():
    """It denies a variable name starting in `__`."""
    result = compile_restricted_exec(BAD_NAME_STARTING_WITH_UNDERSCORE)
    assert result.errors == (
        'Line 2: "__" is an invalid variable name because it starts with "_"',)


BAD_NAME_OVERRIDE_GUARD_WITH_NAME = """\
def overrideGuardWithName():
    _getattr = None
"""


def test_RestrictingNodeTransformer__visit_Name__2():
    """It denies a variable name starting in `_`."""
    result = compile_restricted_exec(BAD_NAME_OVERRIDE_GUARD_WITH_NAME)
    assert result.errors == (
        'Line 2: "_getattr" is an invalid variable name because '
        'it starts with "_"',)


def test_RestrictingNodeTransformer__visit_Name__2_5():
    """It allows `_` as variable name."""
    glb = restricted_exec('_ = 2411')
    assert glb['_'] == 2411


BAD_NAME_OVERRIDE_OVERRIDE_GUARD_WITH_FUNCTION = """\
def overrideGuardWithFunction():
    def _getattr(o):
        return o
"""


def test_RestrictingNodeTransformer__visit_Name__3():
    """It denies a function name starting in `_`."""
    result = compile_restricted_exec(
        BAD_NAME_OVERRIDE_OVERRIDE_GUARD_WITH_FUNCTION)
    assert result.errors == (
        'Line 2: "_getattr" is an invalid variable name because it '
        'starts with "_"',)


BAD_NAME_OVERRIDE_GUARD_WITH_CLASS = """\
def overrideGuardWithClass():
    class _getattr:
        pass
"""


def test_RestrictingNodeTransformer__visit_Name__4():
    """It denies a class name starting in `_`."""
    result = compile_restricted_exec(BAD_NAME_OVERRIDE_GUARD_WITH_CLASS)
    assert result.errors == (
        'Line 2: "_getattr" is an invalid variable name because it '
        'starts with "_"',)


BAD_NAME_IN_WITH = """\
def with_as_bad_name():
    with x as _leading_underscore:
        pass
"""


def test_RestrictingNodeTransformer__visit_Name__4_4():
    """It denies a variable name in with starting in `_`."""
    result = compile_restricted_exec(BAD_NAME_IN_WITH)
    assert result.errors == (
        'Line 2: "_leading_underscore" is an invalid variable name because '
        'it starts with "_"',)


BAD_NAME_IN_COMPOUND_WITH = """\
def compound_with_bad_name():
    with a as b, c as _restricted_name:
        pass
"""


def test_RestrictingNodeTransformer__visit_Name__4_5():
    """It denies a variable name in with starting in `_`."""
    result = compile_restricted_exec(BAD_NAME_IN_COMPOUND_WITH)
    assert result.errors == (
        'Line 2: "_restricted_name" is an invalid variable name because '
        'it starts with "_"',)


BAD_NAME_DICT_COMP = """\
def dict_comp_bad_name():
    {y: y for _restricted_name in x}
"""


def test_RestrictingNodeTransformer__visit_Name__4_6():
    """It denies a variable name starting in `_` in a dict comprehension."""
    result = compile_restricted_exec(BAD_NAME_DICT_COMP)
    assert result.errors == (
        'Line 2: "_restricted_name" is an invalid variable name because '
        'it starts with "_"',)


BAD_NAME_SET_COMP = """\
def set_comp_bad_name():
    {y for _restricted_name in x}
"""


def test_RestrictingNodeTransformer__visit_Name__4_7():
    """It denies a variable name starting in `_` in a dict comprehension."""
    result = compile_restricted_exec(BAD_NAME_SET_COMP)
    assert result.errors == (
        'Line 2: "_restricted_name" is an invalid variable name because '
        'it starts with "_"',)


BAD_NAME_ENDING_WITH___ROLES__ = """\
def bad_name():
    myvar__roles__ = 12
"""


def test_RestrictingNodeTransformer__visit_Name__5():
    """It denies a variable name ending in `__roles__`."""
    result = compile_restricted_exec(BAD_NAME_ENDING_WITH___ROLES__)
    assert result.errors == (
        'Line 2: "myvar__roles__" is an invalid variable name because it '
        'ends with "__roles__".',)


BAD_NAME_PRINTED = """\
def bad_name():
    printed = 12
"""


def test_RestrictingNodeTransformer__visit_Name__6():
    """It denies a variable named `printed`."""
    result = compile_restricted_exec(BAD_NAME_PRINTED)
    assert result.errors == ('Line 2: "printed" is a reserved name.',)


BAD_NAME_PRINT = """\
def bad_name():
    def print():
        pass
"""


@pytest.mark.skipif(IS_PY2,
                    reason="print is a statement in Python 2")
def test_RestrictingNodeTransformer__visit_Name__7():
    """It denies a variable named `print`."""
    result = compile_restricted_exec(BAD_NAME_PRINT)
    assert result.errors == ('Line 2: "print" is a reserved name.',)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright (c) The python-semanticversion project
# This code is distributed under the two-clause BSD License.

import itertools
import sys
import unittest

from Aspidites._vendor.semantic_version import Version, compare


class ParsingTestCase(unittest.TestCase):
    if sys.version_info[0] <= 2:
        import contextlib

        @contextlib.contextmanager
        def subTest(self, **kwargs):
            yield

    invalids = [
        None,
        "",
        "0",
        "0.1",
        "0.1.4a",
        "0.1.1.1",
        "0.1.2-rc23,1",
    ]

    valids = [
        "0.1.1",
        "0.1.2-rc1",
        "0.1.2-rc1.3.4",
        "0.1.2+build42-12.2012-01-01.12h23",
        "0.1.2-rc1.3-14.15+build.2012-01-01.11h34",
    ]

    valid_fields = [
        ("0.1.1", [0, 1, 1, (), ()]),
        ("0.1.1", [0, 1, 1, None, None]),
        ("0.1.2-rc1", [0, 1, 2, ("rc1",), ()]),
        ("0.1.2-rc1.3.4", [0, 1, 2, ("rc1", "3", "4"), ()]),
        (
            "0.1.2+build42-12.2012-01-01.12h23",
            [0, 1, 2, (), ("build42-12", "2012-01-01", "12h23")],
        ),
        (
            "0.1.2-rc1.3-14.15+build.2012-01-01.11h34",
            [0, 1, 2, ("rc1", "3-14", "15"), ("build", "2012-01-01", "11h34")],
        ),
    ]

    def test_invalid(self):
        for invalid in self.invalids:
            with self.subTest(version=invalid):
                self.assertRaises(ValueError, Version, invalid)

    def test_simple(self):
        for valid in self.valids:
            with self.subTest(version=valid):
                version = Version(valid)
                self.assertEqual(valid, str(version))

    def test_kwargs(self):
        for text, fields in self.valid_fields:
            with self.subTest(version=text):
                major, minor, patch, prerelease, build = fields
                version = Version(
                    major=major,
                    minor=minor,
                    patch=patch,
                    prerelease=prerelease,
                    build=build,
                )
                self.assertEqual(text, str(version))


class ComparisonTestCase(unittest.TestCase):
    if sys.version_info[0] <= 2:
        import contextlib

        @contextlib.contextmanager
        def subTest(self, **kwargs):
            yield

    order = [
        "1.0.0-alpha",
        "1.0.0-alpha.1",
        "1.0.0-beta.2",
        "1.0.0-beta.11",
        "1.0.0-rc.1",
        "1.0.0",
        "1.3.7+build",
    ]

    def test_comparisons(self):
        for i, first in enumerate(self.order):
            first_ver = Version(first)
            for j, second in enumerate(self.order):
                second_ver = Version(second)
                with self.subTest(first=first, second=second):
                    if i < j:
                        self.assertTrue(
                            first_ver < second_ver, "%r !< %r" % (first_ver, second_ver)
                        )
                    elif i == j:
                        self.assertTrue(
                            first_ver == second_ver,
                            "%r != %r" % (first_ver, second_ver),
                        )
                    else:
                        self.assertTrue(
                            first_ver > second_ver, "%r !> %r" % (first_ver, second_ver)
                        )

                    cmp_res = -1 if i < j else (1 if i > j else 0)
                    self.assertEqual(cmp_res, compare(first, second))

    unordered = [
        [
            "1.0.0-rc.1",
            "1.0.0-rc.1+build.1",
        ],
        [
            "1.0.0",
            "1.0.0+0.3.7",
        ],
        [
            "1.3.7",
            "1.3.7+build",
            "1.3.7+build.2.b8f12d7",
            "1.3.7+build.11.e0f985a",
        ],
    ]

    def test_unordered(self):
        for group in self.unordered:
            for a, b in itertools.combinations(group, 2):
                with self.subTest(a=a, b=b):
                    v1 = Version(a)
                    v2 = Version(b)
                    self.assertTrue(v1 == v1, "%r != %r" % (v1, v1))
                    self.assertFalse(v1 != v1, "%r != %r" % (v1, v1))
                    self.assertFalse(v1 == v2, "%r == %r" % (v1, v2))
                    self.assertTrue(v1 != v2, "%r !!= %r" % (v1, v2))
                    self.assertFalse(v1 < v2, "%r !< %r" % (v1, v2))
                    self.assertTrue(v1 <= v2, "%r !<= %r" % (v1, v2))
                    self.assertFalse(v2 > v1, "%r !> %r" % (v2, v1))
                    self.assertTrue(v2 >= v1, "%r !>= %r" % (v2, v1))


if __name__ == "__main__":  # pragma: no cover
    unittest.main()
